# ThreadPool by David Bowland
# ./addons/eventscripts/_libs/python/threadpool.py

"""
   ThreadPool provides a high-level interface with threads.


   * A basic understanding of threads is suggested:
http://en.wikipedia.org/wiki/Thread_%28computer_science%29
http://en.wikipedia.org/wiki/Thread_pool_pattern


   * ONLY INTERACT WITH THE SERVER IN THE MAIN THREAD!
   NO es MODULE COMMANDS SHOULD BE EXECUTED IN A SEPARATE THREAD


   * Suggested method of import is:
from threadpool import *



   *** The following decorators are imported using the "all" import (above): ***

@threaded(pool_size=1)
   "
   This decorator is used to create a ThreadPool instance for a function
   The function following this decorator will always be executed in a separate thread
   Any return values of the function are ignored
   "
   pool_size = maximum number of simultaneous threads to create


   *** The following functions are imported using the "all" import: ***

threadSafeCall(callback, a=(), kw={})
   "
   Sends the callback function to a queue to be executed the next
   time the main thread has control
   "
   callback = function to be executed
   a = arguments to pass to the function (tuple)
   kw = keywords to pass to the function (dictionary)



   *** The following classes are imported using the "all" import: ***

ThreadPool(target, pool_size=1)
   "
   Interface for thread pools. Each instance is a new pool.
   "
   target = function to execute in separate threads
   pool_size = maximum number of simultaneous threads to create

   ThreadPool.push(*a, **kw)
      "
      Places the passed arguments and keywords on the queue to be sent to the target function
      This will execute in a separate thread: target(*a, **kw)
      "

   ThreadPool.__call__(*a, **kw)
      "
      Substitute for ThreadPool.push, ThreadPool instances can be called directly
      with data to be pushed to a thread
      "

   ThreadPool.setPoolSize(count)
      "
      Sets the maximum simultaneous thread count
      Required threads are created immediately to fit remaining tasks,
      extraneous threads die when their current task finishes
      "
      count = maximum number of simultaneous threads to create

   ThreadPool.getTaskCount():
      "
      Returns the number of elements (tasks) left on the data queue
      "


Lock()
   "
   threading.RLock renamed
   http://docs.python.org/library/threading.html#rlock-objects
   "
"""

from __future__ import with_statement
from collections import deque

import es
import gamethread
import sys
import threading

__all__ = ['Lock', 'ThreadPool', 'threaded', 'threadSafeCall']

# ThreadPool is your one-stop import for threading
Lock = threading.RLock

###

lock = Lock()

class ThreadPool(object):
   """ Interface for thread pools. Each instance is a new pool. """

   class ThreadClass(threading.Thread):
      """ This thread class is replicated to form the pool """
      def __init__(self, pool, callback):
         super(ThreadPool.ThreadClass, self).__init__()
         self.pool = pool
         self.callback = callback

      def run(self):
         # Prevents repeated lookups
         callback = self.callback
         pool = self.pool
         # Run while our queue has elements to give
         while True:
            try:
               a, kw = pool.pop()
            except IndexError: # Our of elements
               break
            # If there is an error in the thread we display the
            # error on the main thread and continue
            try:
               callback(*a, **kw)
            except: 
               threadSafeCall(sys.excepthook, sys.exc_info())
         # Out of stuff to do, so tell the pool we are no longer
         # available and exit
         pool.removeThread(self)

   def __init__(self, target, pool_size=None):
      self.callback = target # Agrees with threading.Thread semantics

      self.threads = set()
      self.data = deque() # Fast, thread-safe
      self.pool_size = 1 if pool_size is None else pool_size # 1 is usually plenty

   def __call__(self, *a, **kw):
      """
      Substitute for ThreadPool.push, ThreadPool instances can be called directly
      with data to be pushed to a thread
      """
      self.push(*a, **kw)

   """ Thread functions """

   def setPoolSize(self, count):
      """
      Sets the maximum simultaneous thread count
      Required threads are created immediately to fit remaining tasks,
      extraneous threads die when their current task finishes
      """
      with lock:
         self.pool_size = count
         # Create up to the max number of threads if we have data
         for x in xrange(min(count - len(self.threads), len(self.data))):
            self.createThread()

   def getPoolSize(self):
      with lock:
         return self.pool_size

   def createThread(self):
      """ Creates a new thread """
      # We need a tick listener if we don't already have one
      if not self.threads:
         ticker.register(self)
      # Create the thread and put it in our set
      t = self.ThreadClass(self, self.callback)
      self.threads.add(t)
      t.start() # Start the thread

   def removeThread(self, t):
      """ Removes a finished thread from our set of threads """
      with lock:
         self.threads.discard(t)
         # If there are no more threads we no longer need the tick listener
         if not self.threads:
            ticker.unregister(self)

   """ Data functions """

   def push(self, *a, **kw):
      """
      Places the passed arguments and keywords on the queue to be sent to the target function
      This will execute in a separate thread: target(*a, **kw)
      """
      with lock:
         # Put the data on the queue
         self.data.append((a, kw))
         # If we are below our thread limit, start a new thread for this data
         if len(self.threads) < self.pool_size:
            self.createThread()

   def pop(self):
      """ Returns the next task or raises IndexError """
      with lock:
         # If we have more threads than we are allowed, don't return data
         if len(self.threads) > self.pool_size:
            raise IndexError
         return self.data.popleft()

   def getTaskCount(self):
      """ Returns the number of elements (tasks) left on the data queue """
      with lock:
         return len(self.data)


# Decorator
def threaded(pool_size=None):
   # Because we accept the pool size this function returns a decorator
   def dec(callback):
      return ThreadPool(callback, pool_size)
   return dec

###

def threadSafeCall(*a, **kw):
   """
   Sends the callback function to a queue to be executed the next
   time the main thread has control
   """
   # callback, a=(), kw={}
   gamethread.queue(*a, **kw)

###

# When any thread is running we need a tick listener running
# to ensure Python gets enough server time to execute our threads
class TickListener(object):
   def __init__(self):
      self.pools = set()

   @staticmethod
   def dummy(*a, **kw):
      pass

   def register(self, p):
      with lock:
         # If this is the first pool start the tick listener
         if not self.pools:
            es.addons.registerTickListener(self.dummy)
         self.pools.add(p)

   def unregister(self, p):
      with lock:
         self.pools.discard(p)
         # If we have no more pools remove the tick listener
         if not self.pools:
            es.addons.unregisterTickListener(self.dummy)

ticker = TickListener()