# ThreadExample_Friends by David Bowland
# Demonstrates advanced use of the ThreadPool library

from __future__ import with_statement # For URLs
from contextlib import closing # For "with" statement
from xml.etree.ElementTree import XML
from threadpool import * # ThreadPool
from time import time

import es
import re
import urllib2

###

COMMUNITY_URL = 'http://steamcommunity.com/profiles/%s?xml=1'
RE_STEAMID = re.compile(r'^STEAM\_\d\:\d\:\d{3,}$')
TIME_BETWEEN_UPDATES = 4 * 60 * 60 # hr * min * sec

lock = Lock() # Using Lock ensures threads wait their turn when changing data

###

class FriendsList(set):
   lastupdate = 0
   name = ''

   def __init__(self, communityid):
      self.communityid = communityid

   def __str__(self):
      return self.name

#

class FriendsDict(dict):
   def __getitem__(self, item):
      if RE_STEAMID.match(item):
         item = steamIDToCommunityID(item)
      if item in self:
         return super(FriendsDict, self).__getitem__(item)
      x = self[item] = FriendsList(item)
      return x

friends = FriendsDict() # Will hold our friend data

###

# This function will be executed in a separate thread
@threaded()
def getFriends(communityid, name):
   with closing(urllib2.urlopen(COMMUNITY_URL % communityid)) as u:
      data = XML(u.read())

   with lock:
      flist = friends[communityid]
      flist.lastupdate = time()
      # Get the player's community name
      match = flist.name = data.find('customURL')
      if not match is None:
         flist.name = match.text
      # Find the player's friends
      for f in data.find('friends').findall('friend'):
         flist.add(f.find('steamID64').text)
      
   # Pass the result to the callback function
   threadSafeCall(showFriends, (communityid, name))


def steamIDToCommunityID(steamid):
   """ Converts Steam IDs to community IDs """
   x, server, id = steamid.split(':')
   return str(int(id) * 2 + int(server) + 76561197960265728L)

###

def es_player_validated(event_var):
   steamid = event_var['networkid']
   if RE_STEAMID.match(steamid):
      with lock:
         flist = friends[steamid]
         if time() - flist.lastupdate >= TIME_BETWEEN_UPDATES:
            getFriends(flist.communityid, event_var['name'])
         else:
            showFriends(flist.communityid, event_var['name'])


def showFriends(communityid, name):
   def iterFriends():
      for userid in es.getUseridList():
         try:
            if steamIDToCommunityID(es.getplayersteamid(userid)) in flist:
               yield es.getplayername(userid)
         except ValueError: pass

   with lock:
      flist = friends[communityid]
      es.msg('#multi', '#defaultIn-game name: #lightgreen%s#darkgreen  -  #defaultCommunity name: #lightgreen%s' %
       (name, flist.name))

      names = list(iterFriends())
      if len(names) > 1:
         es.msg('#multi', '#defaultCommunity friends:#lightgreen ' +
          '#default,#lightgreen '.join(names[:~0]) + ' #defaultand#lightgreen ' + names[~0])
      elif len(names):
         es.msg('#multi', '#defaultCommunity friends:#lightgreen ' + names[0])